﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;

namespace STC89C5xProgrammer
{
    public partial class Form1 : Form
    {
        public byte[] memory; // array represents memory to be flashed to target, values not specified in hex file will be set to zero
        public ChipInfo ciForm = null;

        private string fileName = "";
        SerialPortForm sForm = null;

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            comPortComboBox.Items.AddRange(SerialPort.GetPortNames());
            comPortComboBox.SelectedIndex = comPortComboBox.FindStringExact(Properties.Settings.Default.comPort);
            fileName = Properties.Settings.Default.fileName;
            highSpeedCheckBox.Checked = Properties.Settings.Default.highSpeed;
            if (fileName.Length > 0)
            {
                filePathLabel.Text = fileName;
                loadFile();
            }

            sForm = new SerialPortForm(this);
            ciForm = new ChipInfo(this);
            // byte[] data = sForm.createBaudData(true, 11059200, 19200);
        }

        private void comPortComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            Properties.Settings.Default.comPort = comPortComboBox.Text;
            Properties.Settings.Default.Save();
        }

        private void filePathLabel_Click(object sender, EventArgs e)
        {
            openFileDialog1.FileName = fileName;
            openFileDialog1.Filter = "Hex files (*.hex)|*.hex|All files (*.*)|*.*";
            openFileDialog1.FilterIndex = 1;
            DialogResult result = openFileDialog1.ShowDialog();
            if (result == DialogResult.OK)
            {
                fileName = openFileDialog1.FileName;
                filePathLabel.Text = fileName;
                Properties.Settings.Default.fileName = fileName;
                Properties.Settings.Default.Save();
                loadFile();
            }
        }

        private void loadFile()
        {
            uploadButton.Visible = false;
            highSpeedCheckBox.Visible = false;

            if (fileName.Length == 0)
            {
                MessageBox.Show("Please choose a file to load", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                fileContentsText.Text = "";
                return;
            }

            if (Path.GetExtension(fileName).ToLower() != ".hex")
            {
                MessageBox.Show("Please choose a hex file (extension .hex)", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                fileContentsText.Text = "";
                return;
            }

            StreamReader sr = null;
            try
            {
                sr = new StreamReader(fileName);
            }
            catch (FileNotFoundException)
            {
                fileContentsText.Text = "";
                return;
            }

            fileContentsText.Text = sr.ReadToEnd();
            sr.Close();

            // test validity of hex file
            int lineNumber = 0;
            int maxAddress = 0;
            foreach (string line in fileContentsText.Lines)
            {
                lineNumber++;

                if (line.Length == 0) continue;

                if (line.Length < 11)
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file is too short", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                if (line[0] != ':')
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file does not start with a colon", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                foreach(char c in line.Substring(1))
                {
                    if (c >= '0' && c <= '9') continue;
                    if (c >= 'A' && c <= 'F') continue;
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file contains non-hex character(s)", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                byte checksum = 0x00;
                for (int i = 1; i < line.Length; i += 2)
                {
                    checksum += Convert.ToByte(line.Substring(i, 2), 16);
                }
                if (checksum != 0x00)
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file has bad checksum", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                int byteCount = Convert.ToByte(line.Substring(1, 2), 16);
                if (line.Length != 2 * byteCount + 11)
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file is wrong length for byte count", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                int recordType = Convert.ToByte(line.Substring(7, 2), 16);
                if (recordType > 1)
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file has record type " + recordType.ToString(), "Hex file is not a I8HEX file", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
                if (recordType == 0)
                {
                    int startAddress = Convert.ToInt32(line.Substring(3, 4), 16);
                    if (startAddress + byteCount - 1 > maxAddress)
                    {
                        maxAddress = startAddress + byteCount - 1;
                    }
                }

            }

            // file seems okay - create memory array containing data to upload to target - round up memory size to multiple of 128
            
            byte[] newMemory = new byte[128 * ((maxAddress + 128) / 128)];
            foreach (string line in fileContentsText.Lines)
            {
                if (line.Length == 0) continue;

                int recordType = Convert.ToByte(line.Substring(7, 2), 16);
                if (recordType == 0)
                {
                    int byteCount = Convert.ToByte(line.Substring(1, 2), 16);
                    int startAddress = Convert.ToInt32(line.Substring(3, 4), 16);
                    for (int i = 0; i < byteCount; i++)
                    {
                        newMemory[startAddress + i] = Convert.ToByte(line.Substring(9 + 2 * i, 2), 16);
                    }

                }
                memory = newMemory;
            }

            // Make upload button visible if a currently available com port has been selected
            foreach (string comPort in comPortComboBox.Items)
            {
                if (comPort == Properties.Settings.Default.comPort)
                {
                    uploadButton.Visible = true;
                    highSpeedCheckBox.Visible = true;
                }
            }
        }

        private void Form1_Resize(object sender, EventArgs e)
        {
            if (this.Width > 170)
            {
                fileContentsText.Width = this.Width - 133;
                progressBar1.Width = this.Width - 133;
            }
            else
            {
                fileContentsText.Width = 37;
                progressBar1.Width = 37;
            }

            if (this.Width > 243)
            {
                filePathLabel.Width = this.Width - 206;
            }
            else
            {
                filePathLabel.Width = 37;
            }

            reloadFileButton.Left = filePathLabel.Width + 111;

            if (this.Width > 452)
            {
                instructionsLabel.Width = Width - 415;
            }
            else
            {
                instructionsLabel.Width = 37;
            }

            if (this.Height > 218)
            {
                fileContentsText.Height = this.Height - 169;
            }
            else
            {
                fileContentsText.Height = 49;
            }
        }

        int timerCount;

        private void timer1_Tick(object sender, EventArgs e)
        {
            int mode = sForm.mode; // snapshot

            if (mode == int.MaxValue) // reading chip info only
            {
                sForm.mode = 0;
                mode = 0;
                instructionsLabel.Visible = false;
                uploadButton.Visible = true;
                highSpeedCheckBox.Visible = true;
                progressBar1.Value = 0;
                timer1.Enabled = false;
                return;
            }

            if (mode > progressBar1.Maximum) mode = progressBar1.Maximum;

            if (timerCount != 0 && mode == 0) // timed out
            {
                instructionsLabel.Visible = false;
                uploadButton.Visible = true;
                highSpeedCheckBox.Visible = true;
                timer1.Enabled = false;
            }
            else if (mode == progressBar1.Maximum)
            {
                if (++timerCount > 60)
                {
                    sForm.mode = 0;
                    mode = 0;
                    instructionsLabel.Visible = false;
                    uploadButton.Visible = true;
                    highSpeedCheckBox.Visible = true;
                    timer1.Enabled = false;
                }
                else if (timerCount > 25)
                {
                    instructionsLabel.Text = "Programming complete!";
                }
            }
            else if (mode > 1)
            {
                instructionsLabel.Text = "Connected!  Uploading to target chip...";
            }
            
            progressBar1.Value = mode;
            if (timerCount == 0) timerCount = 1;
        }

        private void uploadButton_Click(object sender, EventArgs e)
        {
            // sForm.Show();
            sForm.Top = Top + Height;
            sForm.Left = Left;
            sForm.Width = Width;
            instructionsLabel.Text = "Apply power to target chip now!";
            instructionsLabel.Visible = true;
            uploadButton.Visible = false;
            highSpeedCheckBox.Visible = false;
            timerCount = 0;
            timer1.Enabled = true;
            progressBar1.Maximum = memory.Length / 128 + 11;
            ciForm.Show();
            ciForm.Top = Top;
            ciForm.Left = Left + Width;
            ciForm.clear();

            sForm.go(comPortComboBox.Text, 9600, true, highSpeedCheckBox.Checked);
        }

        private void readChipInfoButton_Click(object sender, EventArgs e)
        {
            sForm.Show();
            sForm.Top = Top + Height;
            sForm.Left = Left;
            sForm.Width = Width;
            instructionsLabel.Text = "Apply power to target chip now!";
            instructionsLabel.Visible = true;
            uploadButton.Visible = false;
            highSpeedCheckBox.Visible = false;
            timerCount = 0;
            timer1.Enabled = true;
            if (memory != null)
            {
                progressBar1.Maximum = memory.Length / 128 + 11;
            }
            else
            {
                progressBar1.Maximum = 11;
            }
            ciForm.Show();
            ciForm.Top = Top;
            ciForm.Left = Left + Width;
            ciForm.clear();

            sForm.go(comPortComboBox.Text, 9600, false, highSpeedCheckBox.Checked);

        }

        private void highSpeedCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            Properties.Settings.Default.highSpeed = highSpeedCheckBox.Checked;
            Properties.Settings.Default.Save();
        }

        private void reloadFileButton_Click(object sender, EventArgs e)
        {
            loadFile();
        }
    }
}
